@@ -128,17 +128,41 @@ time_compare (const struct time *a, const struct time *b)
   return 0;
 }
 
-/* Is SEC equal to the current second? */
+/* Is T recent enough that the filesystem could be changed without changing the
+   timestamp again? */
 static bool
-time_is_current (uint64_t sec)
+time_is_current (const struct time *t)
 {
-  static struct timeval cache; /* = { 0, } */
+  static struct time cache; /* = { 0, } */
 
-  /* Cache gettimeofday () results to rule out obviously old time stamps */
-  if ((time_t)sec < cache.tv_sec)
+  struct timeval tv;
+
+  /* This is more difficult than it should be because Linux uses a cheaper time
+     source for filesystem timestamps than for gettimeofday() and they can get
+     slightly out of sync, see
+     https://bugzilla.redhat.com/show_bug.cgi?id=244697 .  This affects even
+     nanosecond timestamps (and don't forget that tv_nsec existence doesn't
+     guarantee that the underlying filesystem has such resolution - it might be
+     microseconds or even coarser).
+
+     The worst case is probably FAT timestamps with 2-second resolution
+     (although using such a filesystem violates POSIX file times requirements).
+
+     So, to be on the safe side, require a >3.0 second difference (2 seconds to
+     make sure the FAT timestamp changed, 1 more to account for the Linux
+     timestamp races).  This large margin might make updatedb marginally more
+     expensive, but it only makes a difference if the directory was very
+     recently updated _and_ is will not be updated again until the next
+     updatedb run; this is not likely to happen for most directories. */
+
+  /* Cache gettimeofday () results to rule out obviously old time stamps;
+     CACHE contains the earliest time we reject as too current. */
+  if (time_compare (t, &cache) < 0)
     return false;
-  gettimeofday (&cache, NULL);
-  return (time_t)sec >= cache.tv_sec;
+  gettimeofday (&tv, NULL);
+  cache.sec = tv.tv_sec - 3;
+  cache.nsec = tv.tv_usec * 1000;
+  return time_compare (t, &cache) >= 0;
 }
 
  /* Directory obstack handling */
@@ -362,10 +386,9 @@ filesystem_is_excluded (const char *path)
 		   sizeof (*conf_prunefs.entries), cmp_string_pointer) != NULL)
 	{
 	  char *dir;
-#ifndef PROC_MOUNTS_PATH
-	  char dbuf[PATH_MAX];
 
-	  dir = realpath (me->mnt_dir, dbuf);
+#ifndef PROC_MOUNTS_PATH
+	  dir = canonicalize_file_name (me->mnt_dir);
 	  if (dir == NULL)
 	    dir = me->mnt_dir;
 #else
@@ -378,10 +401,11 @@ filesystem_is_excluded (const char *path)
 	    /* This is debuging output, don't mark anything for translation */
 	    fprintf (stderr, " => type matches, dir `%s'\n", dir);
 	  if (strcmp (path, dir) == 0)
-	    {
-	      res = true;
-	      goto err_f;
-	    }
+	    res = true;
+	  if (dir != me->mnt_dir)
+	    free(dir);
+	  if (res != false)
+	    goto err_f;
 	}
     }
  err_f:
@@ -647,6 +671,15 @@ scan_cwd (struct directory *dest)
       if (strcmp (de->d_name, ".") == 0 || strcmp (de->d_name, "..") == 0)
 	continue;
       name_size = strlen (de->d_name) + 1;
+      if (name_size == 1)
+	{
+	  /* Unfortunately, this does happen, and mere assert() does not give
+	     users enough information to complain to the right people. */
+	  error (0, 0,
+		 _("file system error: zero-length file name in directory %s"),
+		 dest->path);
+	  continue;
+	}
       assert (name_size > 1);
       entry_size = offsetof (struct entry, name) + name_size;
       if (entry_size > OBSTACK_SIZE_MAX)
@@ -762,10 +795,15 @@ scan (char *path, int *cwd_fd, const struct stat *st_parent,
 	  goto have_dir;
 	}
     }
-  if (dir.time.nsec == 0 && time_is_current (dir.time.sec))
-    /* The directory might be changing right now and we can't be sure the
-       timestamp will yet change, mark the timestamp invalid */
-    dir.time.sec = 0;
+  if (time_is_current (&dir.time))
+    {
+      /* The directory might be changing right now and we can't be sure the
+	 timestamp will be changed again if more changes happen very soon, mark
+	 the timestamp as invalid to force rescanning the directory next time
+	 updatedb is run. */
+      dir.time.sec = 0;
+      dir.time.nsec = 0;
+    }
   did_chdir = true;
   if (safe_chdir (cwd_fd, relative, &st) != 0)
     goto err_chdir;
@@ -942,7 +980,7 @@ main (int argc, char *argv[])
   textdomain (PACKAGE_NAME);
   conf_prepare (argc, argv);
   if (conf_prune_bind_mounts != false)
-    bind_mount_init ();
+    bind_mount_init (MOUNTINFO_PATH);
   lock_file_fd = old_db_open ();
   if (lock_file_fd != -1)
     {
@@ -958,7 +996,6 @@ main (int argc, char *argv[])
 	    error (EXIT_FAILURE, 0,
 		   _("`%s' is locked (probably by an earlier updatedb)"),
 		   conf_output);
-	  fprintf (stderr, "!!%d\n", lock_file_fd);
 	  error (EXIT_FAILURE, errno, _("can not lock `%s'"), conf_output);
 	}
     }
